@@ -5,6 +5,8 @@ h2/stream
 
 An implementation of a HTTP/2 stream.
 """
+import warnings
+
 from enum import Enum, IntEnum
 from hyperframe.frame import (
     HeadersFrame, ContinuationFrame, DataFrame, WindowUpdateFrame,
@@ -15,12 +17,12 @@ from .errors import STREAM_CLOSED
 from .events import (
     RequestReceived, ResponseReceived, DataReceived, WindowUpdated,
     StreamEnded, PushedStreamReceived, StreamReset, TrailersReceived,
-    PriorityUpdated,
+    InformationalResponseReceived,
 )
 from .exceptions import (
     ProtocolError, StreamClosedError, InvalidBodyLengthError
 )
-from .utilities import guard_increment_window
+from .utilities import guard_increment_window, is_informational_response
 
 
 class StreamState(IntEnum):
@@ -40,15 +42,15 @@ class StreamInputs(Enum):
     SEND_DATA = 3
     SEND_WINDOW_UPDATE = 4
     SEND_END_STREAM = 5
-    SEND_PRIORITY = 6
-    RECV_HEADERS = 7
-    RECV_PUSH_PROMISE = 8
-    RECV_RST_STREAM = 9
-    RECV_DATA = 10
-    RECV_WINDOW_UPDATE = 11
-    RECV_END_STREAM = 12
-    RECV_PRIORITY = 13
-    RECV_CONTINUATION = 14  # Added in 2.0.0
+    RECV_HEADERS = 6
+    RECV_PUSH_PROMISE = 7
+    RECV_RST_STREAM = 8
+    RECV_DATA = 9
+    RECV_WINDOW_UPDATE = 10
+    RECV_END_STREAM = 11
+    RECV_CONTINUATION = 12  # Added in 2.0.0
+    SEND_INFORMATIONAL_HEADERS = 13  # Added in 2.2.0
+    RECV_INFORMATIONAL_HEADERS = 14  # Added in 2.2.0
 
 
 # This array is initialized once, and is indexed by the stream states above.
@@ -285,6 +287,28 @@ class H2StreamStateMachine(object):
         assert previous_state == StreamState.CLOSED
         raise StreamClosedError(self.stream_id)
 
+    def push_on_closed_stream(self, previous_state):
+        """
+        Called when an attempt is made to push on an already-closed stream.
+
+        This essentially overrides the standard logic by providing a more
+        useful error message. It's necessary because simply indicating that the
+        stream is closed is not enough: there is now a new stream that is not
+        allowed to be there. The only recourse is to tear the whole connection
+        down.
+        """
+        assert previous_state == StreamState.CLOSED
+        raise ProtocolError("Attempted to push on closed stream.")
+
+    def recv_informational_response(self, previous_state):
+        """
+        Called when an informational header block is received (that is, a block
+        where the :status header has a 1XX value).
+        """
+        event = InformationalResponseReceived()
+        event.stream_id = self.stream_id
+        return [event]
+
 
 # STATE MACHINE
 #
@@ -369,10 +393,6 @@ _transitions = {
     (StreamState.IDLE, StreamInputs.RECV_PUSH_PROMISE):
         (H2StreamStateMachine.recv_new_pushed_stream,
             StreamState.RESERVED_REMOTE),
-    (StreamState.IDLE, StreamInputs.SEND_PRIORITY):
-        (None, StreamState.IDLE),
-    (StreamState.IDLE, StreamInputs.RECV_PRIORITY):
-        (None, StreamState.IDLE),
 
     # State: reserved local
     (StreamState.RESERVED_LOCAL, StreamInputs.SEND_HEADERS):
@@ -387,10 +407,6 @@ _transitions = {
         (None, StreamState.CLOSED),
     (StreamState.RESERVED_LOCAL, StreamInputs.RECV_RST_STREAM):
         (H2StreamStateMachine.stream_reset, StreamState.CLOSED),
-    (StreamState.RESERVED_LOCAL, StreamInputs.SEND_PRIORITY):
-        (None, StreamState.RESERVED_LOCAL),
-    (StreamState.RESERVED_LOCAL, StreamInputs.RECV_PRIORITY):
-        (None, StreamState.RESERVED_LOCAL),
 
     # State: reserved remote
     (StreamState.RESERVED_REMOTE, StreamInputs.RECV_HEADERS):
@@ -406,10 +422,6 @@ _transitions = {
         (None, StreamState.CLOSED),
     (StreamState.RESERVED_REMOTE, StreamInputs.RECV_RST_STREAM):
         (H2StreamStateMachine.stream_reset, StreamState.CLOSED),
-    (StreamState.RESERVED_REMOTE, StreamInputs.SEND_PRIORITY):
-        (None, StreamState.RESERVED_REMOTE),
-    (StreamState.RESERVED_REMOTE, StreamInputs.RECV_PRIORITY):
-        (None, StreamState.RESERVED_REMOTE),
 
     # State: open
     (StreamState.OPEN, StreamInputs.SEND_HEADERS):
@@ -436,10 +448,10 @@ _transitions = {
         (H2StreamStateMachine.send_push_promise, StreamState.OPEN),
     (StreamState.OPEN, StreamInputs.RECV_PUSH_PROMISE):
         (H2StreamStateMachine.recv_push_promise, StreamState.OPEN),
-    (StreamState.OPEN, StreamInputs.SEND_PRIORITY):
-        (None, StreamState.OPEN),
-    (StreamState.OPEN, StreamInputs.RECV_PRIORITY):
+    (StreamState.OPEN, StreamInputs.SEND_INFORMATIONAL_HEADERS):
         (None, StreamState.OPEN),
+    (StreamState.OPEN, StreamInputs.RECV_INFORMATIONAL_HEADERS):
+        (H2StreamStateMachine.recv_informational_response, StreamState.OPEN),
 
     # State: half-closed remote
     (StreamState.HALF_CLOSED_REMOTE, StreamInputs.SEND_HEADERS):
@@ -465,10 +477,6 @@ _transitions = {
             StreamState.HALF_CLOSED_REMOTE),
     (StreamState.HALF_CLOSED_REMOTE, StreamInputs.RECV_PUSH_PROMISE):
         (H2StreamStateMachine.send_reset, StreamState.CLOSED),
-    (StreamState.HALF_CLOSED_REMOTE, StreamInputs.SEND_PRIORITY):
-        (None, StreamState.HALF_CLOSED_REMOTE),
-    (StreamState.HALF_CLOSED_REMOTE, StreamInputs.RECV_PRIORITY):
-        (None, StreamState.HALF_CLOSED_REMOTE),
     (StreamState.HALF_CLOSED_REMOTE, StreamInputs.RECV_CONTINUATION):
         (H2StreamStateMachine.send_reset, StreamState.CLOSED),
 
@@ -491,18 +499,10 @@ _transitions = {
     (StreamState.HALF_CLOSED_LOCAL, StreamInputs.RECV_PUSH_PROMISE):
         (H2StreamStateMachine.recv_push_promise,
             StreamState.HALF_CLOSED_LOCAL),
-    (StreamState.HALF_CLOSED_LOCAL, StreamInputs.SEND_PRIORITY):
-        (None, StreamState.HALF_CLOSED_LOCAL),
-    (StreamState.HALF_CLOSED_LOCAL, StreamInputs.RECV_PRIORITY):
-        (None, StreamState.HALF_CLOSED_LOCAL),
 
     # State: closed
     (StreamState.CLOSED, StreamInputs.RECV_WINDOW_UPDATE):
         (H2StreamStateMachine.window_updated, StreamState.CLOSED),
-    (StreamState.CLOSED, StreamInputs.SEND_PRIORITY):
-        (None, StreamState.CLOSED),
-    (StreamState.CLOSED, StreamInputs.RECV_PRIORITY):
-        (None, StreamState.CLOSED),
     (StreamState.CLOSED, StreamInputs.RECV_RST_STREAM):
         (None, StreamState.CLOSED),  # Swallow further RST_STREAMs
 
@@ -514,7 +514,7 @@ _transitions = {
     (StreamState.CLOSED, StreamInputs.RECV_DATA):
         (H2StreamStateMachine.send_reset, StreamState.CLOSED),
     (StreamState.CLOSED, StreamInputs.RECV_PUSH_PROMISE):
-        (H2StreamStateMachine.send_reset, StreamState.CLOSED),
+        (H2StreamStateMachine.push_on_closed_stream, StreamState.CLOSED),
     (StreamState.CLOSED, StreamInputs.RECV_END_STREAM):
         (None, StreamState.CLOSED),
     (StreamState.CLOSED, StreamInputs.RECV_CONTINUATION):
@@ -524,7 +524,7 @@ _transitions = {
     (StreamState.CLOSED, StreamInputs.SEND_HEADERS):
         (H2StreamStateMachine.send_on_closed_stream, StreamState.CLOSED),
     (StreamState.CLOSED, StreamInputs.SEND_PUSH_PROMISE):
-        (H2StreamStateMachine.send_on_closed_stream, StreamState.CLOSED),
+        (H2StreamStateMachine.push_on_closed_stream, StreamState.CLOSED),
     (StreamState.CLOSED, StreamInputs.SEND_RST_STREAM):
         (H2StreamStateMachine.send_on_closed_stream, StreamState.CLOSED),
     (StreamState.CLOSED, StreamInputs.SEND_DATA):
@@ -586,13 +586,37 @@ class H2Stream(object):
         Returns a list of HEADERS/CONTINUATION frames to emit as either headers
         or trailers.
         """
+        # Convert headers to two-tuples.
+        # FIXME: The fallback for dictionary headers is to be removed in 3.0.
+        try:
+            warnings.warn(
+                "Implicit conversion of dictionaries to two-tuples for "
+                "headers is deprecated and will be removed in 3.0.",
+                DeprecationWarning
+            )
+            headers = headers.items()
+        except AttributeError:
+            headers = headers
+
         # Because encoding headers makes an irreversible change to the header
-        # compression context, we make the state transition *first*.
+        # compression context, we make the state transition before we encode
+        # them.
+
+        # First, check if we're a client. If we are, no problem: if we aren't,
+        # we need to scan the header block to see if this is an informational
+        # response.
+        input_ = StreamInputs.SEND_HEADERS
+        if ((not self.state_machine.client) and
+                is_informational_response(headers)):
+            if end_stream:
+                raise ProtocolError(
+                    "Cannot set END_STREAM on informational responses."
+                )
+
+            input_ = StreamInputs.SEND_INFORMATIONAL_HEADERS
 
         # This does not trigger any events.
-        events = self.state_machine.process_input(
-            StreamInputs.SEND_HEADERS
-        )
+        events = self.state_machine.process_input(input_)
         assert not events
 
         hf = HeadersFrame(self.stream_id)
@@ -708,7 +732,16 @@ class H2Stream(object):
         """
         Receive a set of headers (or trailers).
         """
-        events = self.state_machine.process_input(StreamInputs.RECV_HEADERS)
+        if is_informational_response(headers):
+            if end_stream:
+                raise ProtocolError(
+                    "Cannot set END_STREAM on informational responses"
+                )
+            input_ = StreamInputs.RECV_INFORMATIONAL_HEADERS
+        else:
+            input_ = StreamInputs.RECV_HEADERS
+
+        events = self.state_machine.process_input(input_)
 
         if end_stream:
             events += self.state_machine.process_input(
@@ -789,27 +822,6 @@ class H2Stream(object):
 
         return [], events
 
-    def priority_changed_remote(self, frame):
-        """
-        The remote side of the stream sent priority information.
-        """
-        event = PriorityUpdated()
-        event.stream_id = frame.stream_id
-        event.depends_on = frame.depends_on
-        event.exclusive = frame.exclusive
-
-        # Weight is an integer between 1 and 256, but the byte only allows
-        # 0 to 255: add one.
-        event.weight = frame.stream_weight + 1
-
-        # A stream may not depend on itself.
-        if event.depends_on == self.stream_id:
-            raise ProtocolError(
-                "Stream %d may not depend on itself" % self.stream_id
-            )
-
-        return [event]
-
     def _build_headers_frames(self,
                               headers,
                               encoder,
@@ -817,12 +829,6 @@ class H2Stream(object):
         """
         Helper method to build headers or push promise frames.
         """
-        # Convert headers to two-tuples.
-        try:
-            headers = headers.items()
-        except AttributeError:
-            headers = headers
-
         headers = ((name.lower(), value) for name, value in headers)
         encoded_headers = encoder.encode(headers)
 
